@@ -28,6 +28,9 @@
 #include "gtksourcesnippetchunk-private.h"
 #include "gtksourcesnippetmanager.h"
 #include "gtksourceview-private.h"
+#include "gtksourceutils-private.h"
+
+static void gtk_source_view_snippets_update_informative (GtkSourceViewSnippets *snippets);
 
 static void
 gtk_source_view_snippets_block (GtkSourceViewSnippets *snippets)
@@ -67,11 +70,36 @@ static void
 gtk_source_view_snippets_scroll_to_insert (GtkSourceViewSnippets *snippets)
 {
 	GtkTextMark *mark;
+	GtkTextIter iter;
+	GdkRectangle area;
+	GdkRectangle visible_rect;
+	double x, y;
 
 	g_assert (snippets != NULL);
 
 	mark = gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (snippets->buffer));
-	gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW (snippets->view), mark);
+	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (snippets->buffer), &iter, mark);
+	gtk_text_view_get_iter_location (GTK_TEXT_VIEW (snippets->view), &iter, &area);
+	gtk_text_view_get_visible_rect (GTK_TEXT_VIEW (snippets->view), &visible_rect);
+
+	if (area.x < visible_rect.x)
+		x = area.x;
+	else if (area.x > visible_rect.x + visible_rect.width)
+		x = area.x - visible_rect.width;
+	else
+		x = visible_rect.x;
+
+	if (area.y < visible_rect.y)
+		y = area.y;
+	else if (area.y > visible_rect.y + visible_rect.height)
+		y = area.y - visible_rect.height;
+	else
+		y = visible_rect.y;
+
+	gtk_adjustment_set_value (gtk_scrollable_get_hadjustment (GTK_SCROLLABLE (snippets->view)), x);
+	gtk_adjustment_set_value (gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (snippets->view)), y);
+
+	gtk_source_view_snippets_update_informative (snippets);
 }
 
 static void
@@ -213,8 +241,15 @@ buffer_cursor_moved_cb (GtkSourceBuffer       *buffer,
 
 			snippet = g_queue_peek_head (&snippets->queue);
 		}
-	}
 
+		if (snippet == NULL)
+		{
+			if (snippets->informative != NULL)
+			{
+				gtk_widget_hide (GTK_WIDGET (snippets->informative));
+			}
+		}
+	}
 }
 
 void
@@ -312,22 +347,39 @@ gtk_source_view_snippets_update_informative (GtkSourceViewSnippets *snippets)
 		goto hide_informative;
 	}
 
+	if (snippets->informative == NULL)
+	{
+		snippets->informative = g_object_new (GTK_SOURCE_TYPE_INFORMATIVE,
+		                                      "position", GTK_POS_TOP,
+		                                      "message-type", GTK_MESSAGE_INFO,
+		                                      "icon-name", "completion-snippet-symbolic",
+		                                      NULL);
+		_gtk_source_view_add_assistant (snippets->view,
+		                                GTK_SOURCE_ASSISTANT (snippets->informative));
+	}
+
 	_gtk_source_assistant_set_mark (GTK_SOURCE_ASSISTANT (snippets->informative), chunk->begin_mark);
 	gtk_source_informative_set_message (snippets->informative, tooltip_text);
 
 	if (gtk_widget_get_visible (GTK_WIDGET (snippets->informative)))
 	{
-		_gtk_source_assistant_reposition (GTK_SOURCE_ASSISTANT (snippets->informative));
+		_gtk_source_assistant_set_needs_position (GTK_SOURCE_ASSISTANT (snippets->informative));
 	}
 	else
 	{
-		gtk_widget_show (GTK_WIDGET (snippets->informative));
+		if (gtk_widget_get_mapped (GTK_WIDGET (snippets->view)))
+		{
+			gtk_widget_show (GTK_WIDGET (snippets->informative));
+		}
 	}
 
 	return;
 
 hide_informative:
-	gtk_widget_hide (GTK_WIDGET (snippets->informative));
+	if (snippets->informative != NULL)
+	{
+		gtk_widget_hide (GTK_WIDGET (snippets->informative));
+	}
 }
 
 static void
@@ -343,7 +395,7 @@ gtk_source_view_snippets_notify_position_cb (GtkSourceViewSnippets *snippets,
 }
 
 static void
-gtk_source_view_snippets_bind_cb (GtkSourceSignalGroup  *signals,
+gtk_source_view_snippets_bind_cb (GSignalGroup          *signals,
                                   GtkSourceSnippet      *snippet,
                                   GtkSourceViewSnippets *snippets)
 {
@@ -367,33 +419,25 @@ _gtk_source_view_snippets_init (GtkSourceViewSnippets *snippets,
 	memset (snippets, 0, sizeof *snippets);
 	snippets->view = view;
 
-	snippets->snippet_signals = gtk_source_signal_group_new (GTK_SOURCE_TYPE_SNIPPET);
+	snippets->snippet_signals = g_signal_group_new (GTK_SOURCE_TYPE_SNIPPET);
 
 	g_signal_connect (snippets->snippet_signals,
 	                  "bind",
 	                  G_CALLBACK (gtk_source_view_snippets_bind_cb),
 	                  snippets);
 
-	gtk_source_signal_group_connect_data (snippets->snippet_signals,
-	                                      "notify::focus-position",
-	                                      G_CALLBACK (gtk_source_view_snippets_notify_position_cb),
-	                                      snippets,
-	                                      NULL,
-	                                      G_CONNECT_SWAPPED | G_CONNECT_AFTER);
+	g_signal_group_connect_data (snippets->snippet_signals,
+	                             "notify::focus-position",
+	                             G_CALLBACK (gtk_source_view_snippets_notify_position_cb),
+	                             snippets,
+	                             NULL,
+	                             G_CONNECT_SWAPPED | G_CONNECT_AFTER);
 
 	if (GTK_SOURCE_IS_BUFFER (buffer))
 	{
 		_gtk_source_view_snippets_set_buffer (snippets,
 		                                      GTK_SOURCE_BUFFER (buffer));
 	}
-
-	snippets->informative = g_object_new (GTK_SOURCE_TYPE_INFORMATIVE,
-	                                      "position", GTK_POS_TOP,
-	                                      "message-type", GTK_MESSAGE_INFO,
-	                                      "icon-name", "completion-snippet-symbolic",
-	                                      NULL);
-	_gtk_source_view_add_assistant (snippets->view,
-	                                GTK_SOURCE_ASSISTANT (snippets->informative));
 }
 
 void
@@ -415,12 +459,15 @@ _gtk_source_view_snippets_shutdown (GtkSourceViewSnippets *snippets)
 	g_clear_signal_handler (&snippets->buffer_cursor_moved_handler,
 	                        snippets->buffer);
 
-	_gtk_source_view_remove_assistant (snippets->view,
-	                                   GTK_SOURCE_ASSISTANT (snippets->informative));
+	if (snippets->informative != NULL)
+	{
+		_gtk_source_view_remove_assistant (snippets->view,
+		                                   GTK_SOURCE_ASSISTANT (snippets->informative));
+	}
 
 	if (snippets->snippet_signals != NULL)
 	{
-		gtk_source_signal_group_set_target (snippets->snippet_signals, NULL);
+		g_signal_group_set_target (snippets->snippet_signals, NULL);
 		g_clear_object (&snippets->snippet_signals);
 	}
 
@@ -511,6 +558,7 @@ _gtk_source_view_snippets_key_pressed (GtkSourceViewSnippets *snippets,
 {
 	GdkModifierType modifiers;
 	gboolean editable;
+	gboolean ret = GDK_EVENT_PROPAGATE;
 
 	g_return_val_if_fail (snippets != NULL, FALSE);
 	g_return_val_if_fail (snippets->view != NULL, FALSE);
@@ -520,9 +568,9 @@ _gtk_source_view_snippets_key_pressed (GtkSourceViewSnippets *snippets,
 	 * the view, such as with completion providers.
 	 */
 
-	if (snippets->buffer == NULL)
+	if (snippets->buffer == NULL || snippets->queue.length == 0)
 	{
-		return FALSE;
+		return GDK_EVENT_PROPAGATE;
 	}
 
 	/* Be careful when testing for modifier state equality:
@@ -554,7 +602,8 @@ _gtk_source_view_snippets_key_pressed (GtkSourceViewSnippets *snippets,
 
 				gtk_source_view_snippets_scroll_to_insert (snippets);
 
-				return GDK_EVENT_STOP;
+				ret = GDK_EVENT_STOP;
+				goto cleanup;
 			}
 			else if (state & GDK_SHIFT_MASK)
 			{
@@ -565,7 +614,8 @@ _gtk_source_view_snippets_key_pressed (GtkSourceViewSnippets *snippets,
 
 				gtk_source_view_snippets_scroll_to_insert (snippets);
 
-				return GDK_EVENT_STOP;
+				ret = GDK_EVENT_STOP;
+				goto cleanup;
 			}
 		}
 
@@ -580,11 +630,22 @@ _gtk_source_view_snippets_key_pressed (GtkSourceViewSnippets *snippets,
 		    gtk_source_view_snippets_try_expand (snippets, &end))
 		{
 			gtk_source_view_snippets_scroll_to_insert (snippets);
-			return GDK_EVENT_STOP;
+
+			ret = GDK_EVENT_STOP;
+			goto cleanup;
 		}
 	}
 
-	return GDK_EVENT_PROPAGATE;
+cleanup:
+	if (snippets->queue.length == 0)
+	{
+		if (snippets->informative)
+		{
+			gtk_widget_hide (GTK_WIDGET (snippets->informative));
+		}
+	}
+
+	return ret;
 }
 
 void
@@ -621,7 +682,7 @@ _gtk_source_view_snippets_push (GtkSourceViewSnippets *snippets,
 	}
 	else
 	{
-		gtk_source_signal_group_set_target (snippets->snippet_signals, snippet);
+		g_signal_group_set_target (snippets->snippet_signals, snippet);
 	}
 }
 
@@ -660,12 +721,18 @@ _gtk_source_view_snippets_pop (GtkSourceViewSnippets *snippets)
 		gtk_source_view_snippets_scroll_to_insert (snippets);
 
 		g_object_unref (snippet);
-
-		gtk_widget_hide (GTK_WIDGET (snippets->informative));
 	}
 
 	snippet = g_queue_peek_head (&snippets->queue);
-	gtk_source_signal_group_set_target (snippets->snippet_signals, snippet);
+	g_signal_group_set_target (snippets->snippet_signals, snippet);
+
+	if (snippet == NULL)
+	{
+		if (snippets->informative != NULL)
+		{
+			gtk_widget_hide (GTK_WIDGET (snippets->informative));
+		}
+	}
 }
 
 void
